"use strict";
/*----------------------------------
- DEPENDANCES
----------------------------------*/
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
exports.__esModule = true;
exports.filenameToImportName = void 0;
// Node
var path_1 = __importDefault(require("path"));
var fs_1 = __importDefault(require("fs"));
var generator_1 = __importDefault(require("@babel/generator"));
var micromatch_1 = __importDefault(require("micromatch"));
/*----------------------------------
- WEBPACK RULE
----------------------------------*/
var ruleBuilder = function (options, rules) { return [Plugin, __assign(__assign({}, options), { rules: rules })]; };
module.exports = ruleBuilder;
exports["default"] = ruleBuilder;
/*----------------------------------
- CONST
----------------------------------*/
var LogPrefix = '[babel][glob-imports]';
var MetasPrefix = 'metas:';
var filenameToImportName = function (filename) { return filename.replace(/[^a-z0-9]/gi, '_') || 'index'; };
exports.filenameToImportName = filenameToImportName;
module.exports.filenameToImportName = exports.filenameToImportName;
function uid() {
    var letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz';
    var uniqueId = '';
    for (var i = 0; i < 5; i++) {
        uniqueId += letters.charAt(Math.floor(Math.random() * letters.length));
    }
    return uniqueId;
}
/*----------------------------------
- PLUGIN
----------------------------------*/
function Plugin(babel, options) {
    /*----------------------------------
    - DEFINITION
    ----------------------------------*/
    var t = babel.types;
    var currentFile;
    var plugin = {
        pre: function (state) {
            currentFile = state.opts.filename;
        },
        visitor: {
            // require("@/server/routes/**/*.ts");
            CallExpression: function (instruction) {
                return replaceRequire(instruction);
            },
            // import templates from '@/earn/serveur/emails/*.hbs';
            // import { notifications, inscription } from '@/earn/serveur/emails/*.hbs';
            // import * as templates from '@/earn/serveur/emails/*.hbs';
            ImportDeclaration: function (instruction) {
                return replaceImport(instruction);
            }
        }
    };
    /*----------------------------------
    - REPLACEMENT
    ----------------------------------*/
    function replaceRequire(instruction) {
        var _a = instruction.node, callee = _a.callee, args = _a.arguments;
        if (!(callee.type === "Identifier" && callee.name === 'require'
            &&
                args.length === 1 && args[0].type === "StringLiteral"))
            return;
        // Glob
        var request = {
            type: 'require',
            source: args[0].value,
            from: currentFile,
            withMetas: false
        };
        // Determine if we have to return metas
        if (request.source.startsWith(MetasPrefix)) {
            request.source = request.source.slice(6);
            request.withMetas = true;
        }
        // Find files matched b the glob expression
        var found = findFiles(request);
        if (found === null)
            return;
        var replace = found.replace, files = found.files;
        var replacement;
        // Custom rule
        if (replace !== undefined)
            replacement = replace(request, files, t);
        // Default behavior
        if (replacement === undefined) {
            /* Replace with Metas:
                [
                    {
                        filename: "/server/routes/users/index.ts",
                        matches: ["users", "auth"],
                        exports: require("/server/routes/users/auth.ts")
                    }
                ]
            */
            if (request.withMetas) {
                replacement = [
                    t.expressionStatement(t.arrayExpression(files.map(function (file) { return fileMetasObject(file); })))
                ];
                /*
                    Replace without Metas:
                    [
                        require("/server/routes/users/auth.ts")
                    ]
                */
            }
            else {
                replacement = [
                    t.expressionStatement(t.arrayExpression(files.map(function (file) {
                        // require("/server/routes/users/auth.ts"
                        return t.callExpression(t.identifier('require'), [t.stringLiteral(file.filename)]);
                    })))
                ];
            }
        }
        debugReplacement(instruction, request, found, replacement);
        instruction.replaceWithMultiple(replacement);
    }
    function replaceImport(instruction) {
        // Import request metadata
        var request = getImportRequest(instruction.node);
        // Find files matched b the glob expression
        var found = findFiles(request);
        if (found === null)
            return;
        var replace = found.replace, files = found.files;
        var replacement;
        // Custom replacement rule (defined in the plugin options)
        if (replace !== undefined)
            replacement = replace(request, files, t);
        // Default replacement rule
        if (replacement === undefined) {
            // Import all the files
            var importations = importfiles(files, request);
            // Simply import files
            if (request.imported === undefined) {
                replacement = __spreadArray([], __read(importations.declarations), false);
                // Combine the imported files in one object
            }
            else if (request.imported.type === 'default' || request.imported.type === 'all') {
                var importedfiles = importations.files.map(function (file) { return t.objectProperty(t.stringLiteral(file.exportName), request.withMetas
                    ? fileMetasObject(file, t.identifier(file.local))
                    : t.identifier(file.local)); });
                replacement = __spreadArray(__spreadArray([], __read(importations.declarations), false), [
                    t.variableDeclaration('const', [
                        t.variableDeclarator(t.identifier(request.imported.name), t.objectExpression(importedfiles))
                    ])
                ], false);
            }
        }
        debugReplacement(instruction, request, found, replacement);
        instruction.replaceWithMultiple(replacement);
    }
    function debugReplacement(instruction, request, found, replacement) {
        var _a, _b;
        var hasDebugComment = (_a = instruction.node.leadingComments) === null || _a === void 0 ? void 0 : _a.find(function (comment) { return comment.value === ' @babel-debug'; });
        var debugWhat;
        if (hasDebugComment)
            debugWhat = 'import instruction';
        else if ((_b = found.transformer) === null || _b === void 0 ? void 0 : _b.debug)
            debugWhat = 'transformer' + (found.transformer.name ? ' ' + found.transformer.name : '');
        else if (options.debug)
            debugWhat = 'glob import plugin';
        else
            return;
        console.log(LogPrefix, "Debug " + debugWhat + " in file " + currentFile + '.', dbgTitle('Original import:'), (0, generator_1["default"])(instruction.node).code, dbgTitle('Parsed import:'), request, dbgTitle('Found files:'), found.files, dbgTitle('Generated import:'), replacement ? (0, generator_1["default"])(t.program(replacement)).code : 'nothing');
    }
    function dbgTitle(title) {
        return '\n------- ' + title + ' ----------------------\n';
    }
    function getImportRequest(node) {
        var e_1, _a;
        var imported = undefined;
        try {
            for (var _b = __values(node.specifiers), _c = _b.next(); !_c.done; _c = _b.next()) {
                var specifier = _c.value;
                /*
                    import templates from '@/earn/serveur/emails/*.hbs';
                    =>
                    import templates_notifications from '@/earn/serveur/emails/notifications.hbs';
                    import templates_inscription from '@/earn/serveur/emails/inscription.hbs';
                    const templates = {
                        notifications: templates_notifications,
                        inscription: templates_inscription,
                    }
                */
                if (specifier.type === 'ImportDefaultSpecifier') {
                    imported = {
                        type: 'default',
                        name: specifier.local.name
                    };
                    /*
                        import { notifications, inscription } from '@/earn/serveur/emails/*.hbs';
                        =>
                        import notifications from '@/earn/serveur/emails/notifications.hbs';
                        import inscription from '@/earn/serveur/emails/inscription.hbs';
                    */
                }
                else if (specifier.type === 'ImportSpecifier') {
                    imported = {
                        type: 'destructuration',
                        names: [specifier.local.name]
                    };
                    /*
                        import * as templates from '@/earn/serveur/emails/*.hbs';
                        =>
                        import * as templates_notifications from '@/earn/serveur/emails/notifications.hbs';
                        import * as templates_inscription from '@/earn/serveur/emails/inscription.hbs';
                        const templates = {
                            notifications: templates_notifications,
                            inscription: templates_inscription,
                        }
                    */
                }
                else if (specifier.type === 'ImportNamespaceSpecifier') {
                    imported = {
                        type: 'all',
                        name: specifier.local.name
                    };
                }
                else {
                    console.warn(LogPrefix, "Unsupported import specifier type:", specifier["type"]);
                }
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b["return"])) _a.call(_b);
            }
            finally { if (e_1) throw e_1.error; }
        }
        // Create import request
        var request = {
            type: 'import',
            imported: imported,
            source: node.source.value,
            from: currentFile,
            withMetas: false
        };
        // Determine if we have to return metas
        if (request.source.startsWith(MetasPrefix)) {
            request.source = request.source.slice(6);
            request.withMetas = true;
        }
        return request;
    }
    function importfiles(files, request) {
        var e_2, _a;
        var _b;
        var importedFiles = [];
        var importDeclarations = [];
        var filePrefix = ((_b = request.imported) === null || _b === void 0 ? void 0 : _b['name']) || uid;
        try {
            for (var files_1 = __values(files), files_1_1 = files_1.next(); !files_1_1.done; files_1_1 = files_1.next()) {
                var file = files_1_1.value;
                /// Exclusion du file actuel
                if (file.filename === currentFile)
                    continue;
                // Import specifiers
                var importSpecifier = void 0;
                var imported = request.imported;
                if (imported !== undefined) {
                    // Création nom d'importation via le nom du file
                    var posSlash = file.filename.lastIndexOf('/') + 1;
                    var posExt = file.filename.lastIndexOf('.');
                    var nomFichier = file.filename.substring(posSlash, posExt > posSlash ? posExt : undefined);
                    var nomFichierPourImport = __spreadArray([filePrefix], __read(file.matches), false).join('_');
                    var exportName = file.matches.join('_');
                    // import templates from '@/earn/serveur/emails/*.hbs';
                    if (imported.type === 'default') {
                        var importName = (0, exports.filenameToImportName)(nomFichierPourImport);
                        importedFiles.push(__assign(__assign({}, file), { imported: nomFichierPourImport, local: importName, exportName: exportName }));
                        importSpecifier = t.importDefaultSpecifier(t.identifier(importName));
                        // import * as templates from '@/earn/serveur/emails/*.hbs';
                    }
                    else if (imported.type === 'all') {
                        var importName = (0, exports.filenameToImportName)(nomFichierPourImport);
                        importedFiles.push(__assign(__assign({}, file), { imported: nomFichierPourImport, local: importName, exportName: exportName }));
                        importSpecifier = t.importNamespaceSpecifier(t.identifier(importName));
                        // import { notifications, inscription } from '@/earn/serveur/emails/*.hbs';
                    }
                    else if (imported.type === 'destructuration' && imported.names.includes(nomFichier)) {
                        importSpecifier = t.importDefaultSpecifier(t.identifier(nomFichierPourImport));
                    }
                    else
                        continue;
                }
                // Création de l'importation
                importDeclarations.push(t.importDeclaration(importSpecifier ? [importSpecifier] : [], t.stringLiteral(file.filename)));
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (files_1_1 && !files_1_1.done && (_a = files_1["return"])) _a.call(files_1);
            }
            finally { if (e_2) throw e_2.error; }
        }
        return { files: importedFiles, declarations: importDeclarations };
    }
    /*----------------------------------
    - FILES
    ----------------------------------*/
    function findFiles(request) {
        var e_3, _a;
        var containsGlob = request.source.includes('*');
        var matchingRule = options.rules.find(function (_a) {
            var test = _a.test, globOnly = _a.globOnly;
            return (test(request)
                &&
                    (globOnly === false
                        ||
                            containsGlob));
        });
        // Nothing to process here
        if (!matchingRule && !containsGlob)
            return null;
        var debugRule = options.debug || (matchingRule === null || matchingRule === void 0 ? void 0 : matchingRule.debug);
        var cheminGlob = request.source;
        // Chemin relatif => Transformation en absolu
        if (cheminGlob[0] === '.')
            cheminGlob = path_1["default"].resolve(path_1["default"].dirname(request.from), request.source);
        // Chemin absolu => Remplacement alias
        else if (options.removeAliases !== undefined)
            cheminGlob = options.removeAliases(request.source);
        // If glob, list files in the search directory
        var matchedFiles = [];
        var wildcardPos = cheminGlob.indexOf('*');
        if (wildcardPos !== -1) {
            var rootDir = cheminGlob.substring(0, wildcardPos);
            var allfiles = getFiles(rootDir);
            // Find matches + keep captured groups
            debugRule && console.log(LogPrefix, "Searching for files matching ".concat(request.source, " in directory ").concat(rootDir));
            var regex = micromatch_1["default"].makeRe(cheminGlob, { capture: true });
            try {
                for (var allfiles_1 = __values(allfiles), allfiles_1_1 = allfiles_1.next(); !allfiles_1_1.done; allfiles_1_1 = allfiles_1.next()) {
                    var file = allfiles_1_1.value;
                    var matches = file.match(regex);
                    if (matches) {
                        var mergedMatches = mergeGlobMatches(matches);
                        matchedFiles.push({ filename: file, matches: mergedMatches });
                    }
                }
            }
            catch (e_3_1) { e_3 = { error: e_3_1 }; }
            finally {
                try {
                    if (allfiles_1_1 && !allfiles_1_1.done && (_a = allfiles_1["return"])) _a.call(allfiles_1);
                }
                finally { if (e_3) throw e_3.error; }
            }
            debugRule && console.log(LogPrefix, 'IMPORT GLOB', request.source, '=>', cheminGlob, matchingRule ? 'from rule' : '', matchedFiles);
        }
        return {
            transformer: matchingRule,
            files: matchedFiles,
            replace: matchingRule === null || matchingRule === void 0 ? void 0 : matchingRule.replace
        };
    }
    function mergeGlobMatches(_a) {
        var _b = __read(_a), fileName = _b[0], matches = _b.slice(1);
        if (matches.length === 0)
            return [];
        var results = [];
        var current = matches[0];
        var currentIndex = current ? fileName.indexOf(current) : -1;
        for (var i = 1; i < matches.length; i++) {
            var next = matches[i];
            if (current === undefined) {
                current = next;
                currentIndex = current ? fileName.indexOf(current) : -1;
                continue;
            }
            if (next === undefined)
                continue;
            var nextIndex = fileName.indexOf(next, currentIndex + current.length);
            if (nextIndex <= currentIndex + current.length) {
                // If the next match is adjacent or overlapping, merge it
                var overlap = currentIndex + current.length - nextIndex;
                current += next.slice(overlap);
            }
            else {
                results.push(current);
                current = next;
                currentIndex = fileName.indexOf(current);
            }
        }
        if (current !== undefined) {
            results.push(current);
        }
        return results;
    }
    function getFiles(dir) {
        var e_4, _a;
        var files = [];
        var dirents = fs_1["default"].readdirSync(dir, { withFileTypes: true });
        try {
            for (var dirents_1 = __values(dirents), dirents_1_1 = dirents_1.next(); !dirents_1_1.done; dirents_1_1 = dirents_1.next()) {
                var dirent = dirents_1_1.value;
                var res = path_1["default"].resolve(dir, dirent.name);
                if (dirent.isDirectory()) {
                    files.push.apply(files, __spreadArray([], __read(getFiles(res)), false));
                }
                else {
                    files.push(res);
                }
            }
        }
        catch (e_4_1) { e_4 = { error: e_4_1 }; }
        finally {
            try {
                if (dirents_1_1 && !dirents_1_1.done && (_a = dirents_1["return"])) _a.call(dirents_1);
            }
            finally { if (e_4) throw e_4.error; }
        }
        return files;
    }
    function fileMetasObject(file, importedIdentifier) {
        return t.objectExpression([
            // filename: "/server/routes/users/auth.ts",
            t.objectProperty(t.identifier('filename'), t.stringLiteral(file.filename)),
            // matches: ["users", "auth"],
            t.objectProperty(t.identifier('matches'), t.arrayExpression(file.matches.map(function (match) { return match === undefined
                ? t.identifier('undefined')
                : t.stringLiteral(match); }))),
            // importedIdentifier === undefined
            // exports: require("/server/routes/users/auth.ts")
            // importedIdentifier !== undefined
            // exports: headhunter_auth
            t.objectProperty(t.identifier('exports'), importedIdentifier === undefined
                ? t.callExpression(t.identifier('require'), [t.stringLiteral(file.filename)])
                : importedIdentifier),
        ]);
    }
    return plugin;
}
